from quicktions import Fraction
from abc import ABC, abstractmethod
from typing import Union


class _VideoSource(ABC):
    """
    Abstract class that is a media source
    containing video information.
    """

    @property
    @abstractmethod
    def copy(
        self
    ):
        """
        Get a copy of this video source instance.
        """
        pass

    @property
    @abstractmethod
    def duration(
        self
    ):
        """
        The duration of the source.
        """
        pass

    @property
    @abstractmethod
    def size(
        self
    ):
        """
        The size of the original source, that can
        match or match not the 'output_size'.
        """
        pass

    def __init__(
        self
    ):
        pass

    @abstractmethod
    def get_video_frame_at_t(
        self,
        t: Union[int, float, Fraction],
        size: Union[tuple[int, int], None] = None,
        do_apply_filters: bool = True
        # TODO: Return 'VideoFrame' or 'numpy' (?)
    ):
        """
        Get the video frame that must be displayed
        at the 't' time moment.
        """
        pass

    @abstractmethod
    def get_audio_frames_at_t(
        self,
        t: Union[int, float, Fraction],
        video_fps: Union[int, float, Fraction],
        do_apply_filters: bool = True
        # TODO: Return 'VideoFrame' or 'numpy' (?)
    ):
        """
        Get the audio frames that must be played
        during the 't' time moment of a video.
        """
        pass

    # TODO: audio frames t (?)
        
class _AudioSource(ABC):
    """
    Abstract class that is a media source
    containing audio information.
    """

    @property
    @abstractmethod
    def copy(
        self
    ):
        """
        Get a copy of this video source instance.
        """
        pass

    @property
    @abstractmethod
    def duration(
        self
    ):
        """
        The duration of the source.
        """
        pass

    def __init__(
        self
    ):
        pass

    @abstractmethod
    def get_audio_frames_at_t(
        self,
        t: Union[int, float, Fraction],
        video_fps: Union[int, float, Fraction],
        do_apply_filters: bool = True
        # TODO: Return 'VideoFrame' or 'numpy' (?)
    ):
        """
        Get the audio frames that must be played
        during the 't' time moment of a video.
        """
        pass




from functools import wraps


def with_t_adjusted_to_media(
    method
):
    """
    Get the real 't' time moment based on the
    video 'start' and 'end'. If they were 
    asking for the t=0.5s but our video was
    subclipped to [1.0, 2.0), the 0.5s must be
    actually the 1.5s of the video because of
    the subclipped time range.

    The formula:
    - `t + self.start`
    """
    @wraps(method)
    def wrapper(
        self,
        t,
        *args,
        **kwargs
    ):
        t += self.start
    
        print(f'The video/audio real t is {str(float(t))}')
        if t >= self.end:
            raise Exception(f'The "t" ({str(t)}) provided is out of range. This video/audio lasts from [{str(self.start)}, {str(self.end)}).')
        
        return method(self, t, *args, **kwargs)
    
    return wrapper



"""
The video track module.
"""
from yta_editor.track.parts.video import _VideoPart
from yta_editor.track.media.video import VideoOnTrack
from yta_editor.track.abstract import _TrackWithAudio, _TrackWithVideo


class VideoTrack(_TrackWithVideo, _TrackWithAudio):
    """
    Class to represent a track in which we place
    videos to build a video project.
    """

    def __init__(
        self,
        index: int,
        size: tuple[int, int],
        fps: float,
        audio_fps: float,
        audio_samples_per_frame: int,
        audio_layout: str,
        audio_format: str
    ):
        # super().__init__(
        #     index = index,
        #     size = size,
        #     fps = fps,
        #     audio_fps = audio_fps,
        #     audio_samples_per_frame = audio_samples_per_frame,
        #     audio_layout = audio_layout,
        #     audio_format = audio_format
        # )
        _TrackWithVideo.__init__(
            self,
            index = index,
            size = size,
            fps = fps
        )
        _TrackWithAudio.__init__(
            self,
            index = index,
            fps = fps,
            audio_fps = audio_fps,
            audio_samples_per_frame = audio_samples_per_frame,
            audio_layout = audio_layout,
            audio_format = audio_format
        )
        
    def _make_part(
        self,
        **kwargs
    ):
        return _VideoPart(**kwargs)
    
    def _make_media_on_track(
        self,
        **kwargs
    ):
        return VideoOnTrack(**kwargs)


from typing import Union


def _update_transparency(
    transparency: Union[float, None],
    format: Union[str, None] = None
) -> Union[float, None]:
    """
    *For internal use only*

    Update the provided 'transparency' value,
    that must be a value in the [0.0, 1.0] range,
    if necessary.

    If the 'format' is provided:
    - Alpha format will force an opaque alpha
    transparency if it is None
    - Non-alpha format will force a None value
    if transparency is provided
    """
    return (
        # Alpha format, no transparency => opaque
        0.0
        if (
            transparency is None and
            format is not None and
            'a' in format 
        ) else
        # Non-alpha format but transparency => None
        None
        if (
            transparency is not None and
            format is not None and
            'a' not in format
        ) else
        # Lower than 0.0 limit
        0.0
        if (
            transparency is not None and
            transparency < 0.0
        ) else
        # Greater than 1.0 limit
        1.0
        if (
            transparency is not None and
            transparency > 1.0
        ) else
        transparency
    )


from yta_editor.track.media.abstract import _MediaOnTrackWithAudio
from yta_editor.media.audio import AudioFileMedia
from yta_validation.parameter import ParameterValidator
from quicktions import Fraction
from typing import Union


class AudioOnTrack(_MediaOnTrackWithAudio):
    """
    An audio media but positioned on an audio
    track in the editor timeline.
    """

    def __init__(
        self,
        media: AudioFileMedia,
        start: Union[int, float, Fraction] = 0.0
    ):
        ParameterValidator.validate_mandatory_instance_of('media', media, AudioFileMedia)

        super().__init__(
            media = media,
            start = start
        )




"""
The parts module.

A part is an element that has been placed
in an specific track. We also consider a
part some imaginary things that are 
fulfilling the track regions in which the
user didn't put any media, so we can 
interact with those 'empty' parts to 
obtain 'empty' audio and video frames.
"""
# from yta_editor.track.parts.audio import _AudioPart
# from yta_editor.track.parts.video import _VideoPart
# from yta_editor.track.parts.abstract import NON_LIMITED_EMPTY_PART_END


# __all__ = [
#     '_AudioPart',
#     '_VideoPart',
#     'NON_LIMITED_EMPTY_PART_END'
# ]


"""
The video frames must be built using the
(height, width) size when giving the numpy
array that will be used for it. We will
receive the values as (width, height) but
we will invert them when needed.

The frames that come from an empty part
are flagged with the .metadata attribute
'is_from_empty_part' so we can recognize
them and ignore when combining on the
timeline. We have that metadata in the
wrapper class we created.

TODO: Check because we have a similar
module in other project or projects.
"""
from yta_editor.utils.alpha import _update_transparency
from yta_numpy.rgba.generator import RGBAFrameGenerator
from av.video.frame import VideoFrame
from typing import Union

import numpy as np


class _BackgroundFrameGenerator:
    """
    Internal class to simplify the way we
    access to the generation of background
    frames form the general generator class.
    """

    @staticmethod
    def full_black(
        size: tuple[int, int] = (1920, 1080),
        dtype: np.dtype = np.uint8,
        # TODO: I disable format by now because I
        # make it dynamic according to the 
        # 'transparency' value provided
        #format: str = 'rgb24',
        pts: Union[int, None] = None,
        time_base: Union['Fraction', None] = None,
        transparency: Union[float, None] = None
    ) -> VideoFrame:
        """
        Get a video frame that is completely black
        and of the given 'size'.

        The 'transparency' must be a float between
        0.0 (opaque) and 1.0 (transparent) to be
        set, or None if you don't want transparency.

        Providing 'transparency' as None will 
        result in a numpy with only 3 dimensions.
        """
        return _BackgroundFrameGenerator._full_color(
            color = 'black',
            size = size,
            dtype = dtype,
            pts = pts,
            time_base = time_base,
            transparency = transparency
        )
    
    @staticmethod
    def full_white(
        size: tuple[int, int] = (1920, 1080),
        dtype: np.dtype = np.uint8,
        # TODO: I disable format by now because I
        # make it dynamic according to the 
        # 'transparency' value provided
        #format: str = 'rgb24',
        pts: Union[int, None] = None,
        time_base: Union['Fraction', None] = None,
        transparency: Union[float, None] = None
    ) -> VideoFrame:
        """
        Get a video frame that is completely white
        and of the given 'size'.

        The 'transparency' must be a float between
        0.0 (opaque) and 1.0 (transparent) to be
        set, or None if you don't want transparency.

        Providing 'transparency' as None will 
        result in a numpy with only 3 dimensions.
        """
        return _BackgroundFrameGenerator._full_color(
            color = 'white',
            size = size,
            dtype = dtype,
            pts = pts,
            time_base = time_base,
            transparency = transparency
        )

    @staticmethod
    def full_red(
        size: tuple[int, int] = (1920, 1080),
        dtype: np.dtype = np.uint8,
        # TODO: I disable format by now because I
        # make it dynamic according to the 
        # 'transparency' value provided
        #format: str = 'rgb24',
        pts: Union[int, None] = None,
        time_base: Union['Fraction', None] = None,
        transparency: Union[float, None] = None
    ) -> VideoFrame:
        """
        Get a video frame that is completely red
        and of the given 'size'.

        The 'transparency' must be a float between
        0.0 (opaque) and 1.0 (transparent) to be
        set, or None if you don't want transparency.

        Providing 'transparency' as None will 
        result in a numpy with only 3 dimensions.
        """
        return _BackgroundFrameGenerator._full_color(
            color = 'red',
            size = size,
            dtype = dtype,
            pts = pts,
            time_base = time_base,
            transparency = transparency
        )
    
    # TODO: Refactor all this please
    @staticmethod
    def full_green(
        size: tuple[int, int] = (1920, 1080),
        dtype: np.dtype = np.uint8,
        # TODO: I disable format by now because I
        # make it dynamic according to the 
        # 'transparency' value provided
        #format: str = 'rgb24',
        pts: Union[int, None] = None,
        time_base: Union['Fraction', None] = None,
        transparency: Union[float, None] = None
    ) -> VideoFrame:
        """
        Get a video frame that is completely green
        and of the given 'size'.

        The 'transparency' must be a float between
        0.0 (opaque) and 1.0 (transparent) to be
        set, or None if you don't want transparency.

        Providing 'transparency' as None will 
        result in a numpy with only 3 dimensions.
        """
        return _BackgroundFrameGenerator._full_color(
            color = 'green',
            size = size,
            dtype = dtype,
            pts = pts,
            time_base = time_base,
            transparency = transparency
        )

    @staticmethod
    def full_blue(
        size: tuple[int, int] = (1920, 1080),
        dtype: np.dtype = np.uint8,
        # TODO: I disable format by now because I
        # make it dynamic according to the 
        # 'transparency' value provided
        #format: str = 'rgb24',
        pts: Union[int, None] = None,
        time_base: Union['Fraction', None] = None,
        transparency: Union[float, None] = None
    ) -> VideoFrame:
        """
        Get a video frame that is completely blue
        and of the given 'size'.

        The 'transparency' must be a float between
        0.0 (opaque) and 1.0 (transparent) to be
        set, or None if you don't want transparency.

        Providing 'transparency' as None will 
        result in a numpy with only 3 dimensions.
        """
        return _BackgroundFrameGenerator._full_color(
            color = 'blue',
            size = size,
            dtype = dtype,
            pts = pts,
            time_base = time_base,
            transparency = transparency
        )
    
    @staticmethod
    def _full_color(
        color: str,
        size: tuple[int, int] = (1920, 1080),
        dtype: np.dtype = np.uint8,
        # TODO: I disable format by now because I
        # make it dynamic according to the 
        # 'transparency' value provided
        #format: str = 'rgb24',
        pts: Union[int, None] = None,
        time_base: Union['Fraction', None] = None,
        transparency: Union[float, None] = None
    ):
        """
        *For internal use only*

        Method that dynamically calls the internal
        method to generate the numpy according to
        the 'color' provided, that must fit the
        name of an existing method of the 
        '_FrameGenerator' class.
        """
        format = (
            'rgba'
            if transparency is not None else
            'rgb24'
        )

        transparency = _update_transparency(
            transparency = transparency,
            format = format
        )

        generator_method = getattr(RGBAFrameGenerator.color, color)
        
        # In the RGBFrameGenerator we use alpha
        # in the [0, 255] range
        # alpha is actually the color presence,
        # which is the opposite to transparency
        alpha = (
            int(255 - transparency * 255)
            if transparency is not None else
            None
        )

        frame = generator_method(
            size = size,
            dtype = dtype,
            alpha = alpha
        )

        return numpy_to_video_frame(
            frame = frame,
            format = format,
            pts = pts,
            time_base = time_base
        )

class VideoFrameGenerator:
    """
    Class to wrap the functionality related to
    generating a pyav video frame.

    This class is useful when we need to 
    generate the black background for empty
    parts within the tracks and in other 
    situations.
    """

    background: _BackgroundFrameGenerator = _BackgroundFrameGenerator
    """
    Shortcut to the background generation.
    """

def numpy_to_video_frame(
    frame: np.ndarray,
    format: str = 'rgb24',
    pts: Union[int, None] = None,
    time_base: Union['Fraction', None] = None
) -> VideoFrame:
    """
    Transform the given numpy 'frame' into a
    pyav video frame with the given 'format'
    and also the 'pts' and/or 'time_base' if
    provided.
    """
    frame = VideoFrame.from_ndarray(
        # TODO: What if we want alpha (?)
        array = frame,
        format = format
    )

    if pts is not None:
        frame.pts = pts

    if time_base is not None:
        frame.time_base = time_base

    return frame


"""
Media elements, that are the ones we will
use and place on the tracks, able to access
to the frames, using the different sources,
and to apply the effects we want.
"""
# from yta_editor.media.audio import AudioFileMedia
# from yta_editor.media.video import VideoFileMedia, VideoColorMedia, VideoImageMedia


# __all__ = [
#     'AudioFileMedia',
#     'VideoFileMedia',
#     'VideoColorMedia',
#     'VideoImageMedia',
# ]


from yta_editor.track.media.abstract import _MediaOnTrackWithAudio, _MediaOnTrackWithVideo
from yta_editor.media.video import VideoFileMedia, VideoImageMedia, VideoColorMedia
from yta_validation.parameter import ParameterValidator
from quicktions import Fraction
from typing import Union


class VideoOnTrack(_MediaOnTrackWithAudio, _MediaOnTrackWithVideo):
    """
    A video media but positioned on a video
    track in the editor timeline.
    """

    def __init__(
        self,
        media: Union[VideoFileMedia, VideoImageMedia, VideoColorMedia],
        start: Union[int, float, Fraction] = 0.0
    ):
        ParameterValidator.validate_mandatory_instance_of('media', media, [VideoFileMedia, VideoImageMedia, VideoColorMedia])

        super().__init__(
            media = media,
            start = start
        )


from yta_constants.ffmpeg import FfmpegAudioLayout, FfmpegAudioFormat, FfmpegPixelFormat
from yta_validation.parameter import ParameterValidator
from av.audio.frame import AudioFrame
from av.audio.layout import AudioLayout
from av.audio.stream import AudioStream
from av.video.stream import VideoStream
from av.video.frame import VideoFrame
from quicktions import Fraction
from av import open as av_open
from typing import Union

import numpy as np


class _AlphaUtils:
    """
    General class to include some utils related
    to alpha channel and the pyav library.
    """

    @staticmethod
    def file_has_alpha_layer(
        filename: str
    ) -> bool:
        """
        Check if the video file with the given 'filename'
        has alpha layer or not, that is defined by an
        'a' in the frame format name. This will test the
        first frame of the video and return a result.

        This doesn't mean that there are transparent
        pixels on the video, just the presence of the
        alpha layer.
        """
        return _AlphaUtils.videoframe_has_alpha_layer(next(_VideoUtils.iter_frames(filename)))

    @staticmethod
    def videoframe_has_alpha_layer(
        frame: VideoFrame
    ) -> bool:
        """
        Check if the provided 'frame' pyav VideoFrame
        has alpha layer or not, that is defined by an
        'a' in the frame format name.
        
        This doesn't mean that some pixel is
        transparent, it only means that there is an
        alpha layer, but the frame could be completely
        opaque.

        The code:
        - `'a' in frame.format.name`
        """
        ParameterValidator.validate_mandatory_instance_of('frame', frame, VideoFrame)

        return 'a' in frame.format.name
    
    @staticmethod
    def numpy_videoframe_has_transparent_pixels(
        frame: 'np.ndarray'
    ) -> bool:
        """
        Check if the provided numpy array has alpha
        pixels or not. If the provided numpy array
        doesn't include an alpha layer the result
        will be False.
        """
        return (
            frame.ndim == 3 and
            frame.shape[2] > 3 and
            np.any(frame[..., 3] < 255)
        )
    
    @staticmethod
    def videoframe_has_transparent_pixels(
        frame: VideoFrame
    ) -> bool:
        """
        Check if the provided frame has alpha pixels
        or not.
        """
        return _AlphaUtils.numpy_videoframe_has_transparent_pixels(frame.to_ndarray(format = 'rgba'))

    @staticmethod
    def file_has_transparent_pixels(
        filename: str
    ) -> bool:
        """
        Check if the video with the provided 'filename'
        has some transparent (at least partially) alpha
        pixels by decoding the frames until finding one
        that has some transparency.

        This process is slow because it will decode the
        whole video until finding some transparent pixel
        (if existing).
        """
        for frame in _VideoUtils.iter_frames(filename):
            if not _AlphaUtils.videoframe_has_alpha_layer(frame):
                break
            
            # Check if alpha channel has some transparency
            if _AlphaUtils.videoframe_has_transparent_pixels(frame):
                return True
            
        return False
    
class _VideoFrameUtils:
    """
    *For internal use only*

    Class to wrap functionality related to video
    frames.
    """
    
    def copy(
        video_frame: VideoFrame
    ) -> 'VideoFrame':
        """
        Create a copy of the provided VideoFrame
        instance. This is similar to .copy().
        """
        return VideoFrame.from_ndarray(
            array = video_frame.to_ndarray(format = video_frame.format.name),
            format = video_frame.format.name
        )
    
    @staticmethod
    def combine_videoframes_with_alpha_layer(
        top_frame: VideoFrame,
        bottom_frame: VideoFrame,
        pixel_format: Union[FfmpegPixelFormat, None] = None
    ) -> VideoFrame:
        """
        Combine the provided 'top_frame' and
        'bottom_frame' transforming them into a
        RGBA and uint8 numpy array.

        The pixel format will be the provided as
        the 'pixel_format' parameter, or the one
        the 'top_frame' has.
        """
        # TODO: What if the size is not the same (?)
        # TODO: What if no alpha layer in the frames (?)
        pixel_format = (
            top_frame.format.name
            if pixel_format is None else
            FfmpegPixelFormat.to_enum(pixel_format).value
        )

        return VideoFrame.from_ndarray(
            array = _VideoFrameUtils.combine_numpy_videoframes_with_alpha_layer(
                top_frame = top_frame.to_ndarray(format = FfmpegPixelFormat.RGBA.value),
                bottom_frame = bottom_frame.to_ndarray(format = FfmpegPixelFormat.RGBA.value)
            ),
            format = pixel_format
        )

    @staticmethod
    def combine_numpy_videoframes_with_alpha_layer(
        top_frame: np.ndarray,
        bottom_frame: np.ndarray
    ) -> np.ndarray:
        """
        Combine 2 RGBA frames, that must be (h, w, 4)
        and uint8, with alpha blending. Useful to
        combine video frames with alpha blending.

        This method returns the new combined frame.
        """
        # To float in [0, 1] range
        top_frame_rgb = top_frame[..., :3].astype(np.float32) / 255.0
        bottom_frame_rgb = bottom_frame[..., :3].astype(np.float32) / 255.0
        top_frame_alpha = top_frame[..., 3:].astype(np.float32) / 255.0
        bottom_frame_alpha = bottom_frame[..., 3:].astype(np.float32) / 255.0

        # Alpha blending
        output_alpha = top_frame_alpha + bottom_frame_alpha * (1 - top_frame_alpha)
        # Avoid division by zero
        output_rgb = (top_frame_rgb * top_frame_alpha + bottom_frame_rgb * bottom_frame_alpha * (1 - top_frame_alpha)) / np.clip(output_alpha, 1e-6, 1.0)

        # Back to uint8
        output_frame = np.zeros_like(top_frame, dtype = np.uint8)
        output_frame[..., :3] = (output_rgb * 255).astype(np.uint8)
        output_frame[..., 3:] = (output_alpha * 255).astype(np.uint8)

        return output_frame
    
class _VideoUtils:
    """
    *For internal use only*

    Class to wrap some utils related to video.
    """

    alpha: _AlphaUtils = _AlphaUtils
    """
    Information about the alpha channel.
    """
    frame: _VideoFrameUtils = _VideoFrameUtils
    """
    Utils related to video frames.
    """

    @staticmethod
    def iter_frames(
        filename: str
    ):
        """
        Iterator over the video frames of the file
        with the given 'filename'.
        """
        container = av_open(file = filename)
        video_stream = (
            container.streams.video[0]
            if container.streams.video else
            None
        )

        if video_stream is None:
            raise Exception(f'No video stream found in "{filename}" file.')
        
        for frame in _iter_frames(
            stream = video_stream
        ):
            yield frame

        return
    
class _AudioFormatUtils:
    """
    *For internal use only*

    Class to wrap some utils related to audio
    format.
    """

    @staticmethod
    def to_dtype(
        audio_format: str
    ) -> Union[np.dtype, None]:
        """
        Transform the given 'audio_format' into
        the corresponding numpy dtype value. If
        the 'audio_format' is not accepted this
        method will return None.

        This method must be used when we are
        building the numpy array that will be 
        used to build a pyav audio frame because
        the pyav 'audio_format' need a specific
        np.dtype to be built.

        For example, 's16' will return 'np.int16'
        and 'fltp' will return 'np.float32'.
        """
        return {
            # Integers and signed
            FfmpegAudioFormat.U8.value:   np.uint8,   # unsigned 8-bit
            FfmpegAudioFormat.S16.value:  np.int16,   # signed 16-bit
            FfmpegAudioFormat.S32.value:  np.int32,   # signed 32-bit

            # Floats
            FfmpegAudioFormat.FLT.value:  np.float32, # 32-bit float (interleaved)
            FfmpegAudioFormat.FLTP.value: np.float32, # 32-bit float (planar)
            FfmpegAudioFormat.DBL.value:  np.float64, # 64-bit float (interleaved)
            FfmpegAudioFormat.DBLP.value: np.float64, # 64-bit float (planar)

            # Very deep extensions
            FfmpegAudioFormat.S64.value:  np.int64,   # signed 64-bit
            FfmpegAudioFormat.S64P.value: np.int64,   # signed 64-bit (planar)

            # Integer planars (each channel separate)
            FfmpegAudioFormat.U8P.value:  np.uint8,
            FfmpegAudioFormat.S16P.value: np.int16,
            FfmpegAudioFormat.S32P.value: np.int32,
        }.get(FfmpegAudioFormat.to_enum(audio_format).value, None)

class _AudioFrameUtils:
    """
    *For internal use only*

    Class to wrap some utils related to audio
    frames.
    """

    @staticmethod
    def trim(
        frame: AudioFrame,
        start: Union[int, float, Fraction],
        end: Union[int, float, Fraction],
        time_base: Fraction
    ) -> AudioFrame:
        """
        Trim an audio frame to obtain the part between
        [start, end), that is provided in seconds.
        """
        # (channels, n_samples)
        samples = frame.to_ndarray()  
        n_samples = samples.shape[1]

        # In seconds
        frame_start = frame.pts * float(time_base)
        frame_end = frame_start + (n_samples / frame.sample_rate)

        # Overlapping 
        cut_start = max(frame_start, float(start))
        cut_end = min(frame_end, float(end))

        if cut_start >= cut_end:
            # No overlapping
            return None  

        # To sample indexes
        start_index = int(round((cut_start - frame_start) * frame.sample_rate))
        end_index = int(round((cut_end - frame_start) * frame.sample_rate))

        new_frame = AudioFrame.from_ndarray(
            # end_index is not included: so [start, end)
            array = samples[:, start_index:end_index],
            format = frame.format,
            layout = frame.layout
        )

        # Set needed attributes
        new_frame.sample_rate = frame.sample_rate
        new_frame.time_base = time_base
        new_frame.pts = int(round(cut_start / float(time_base)))

        return new_frame
    
    @staticmethod
    def silent(
        sample_rate: int,
        number_of_samples: int = 1024,
        layout = FfmpegAudioLayout.STEREO.value,
        format = FfmpegAudioFormat.S16.value,
        pts: Union[int, None] = None,
        time_base: Union[Fraction, None] = None
    ) -> AudioFrame:
        """
        Get an audio frame that is completely silent.
        This is useful when we want to fill the empty
        parts of our tracks.
        """
        layout = FfmpegAudioLayout.to_enum(layout).value
        format = FfmpegAudioFormat.to_enum(format).value

        dtype = _AudioFormatUtils.to_dtype(format)

        if dtype is None:
            raise Exception(f'The format "{format}" is not accepted.')

        number_of_channels = len(AudioLayout(layout).channels)

        # TODO: I leave these comments below because
        # I'm not sure what is true and what is not
        # so, until it is more clear... here it is:
        # For packed (or planar) formats we apply:
        # (1, samples * channels). This is the same
        # amount of data but planar, in 1D only
        # TODO: This wasn't in the previous version
        # and it was working, we were sending the
        # same 'number_of_samples' even when 'fltp'
        # that includes the 'p'
        # TODO: This is making the audio last 2x
        # if 'p' in format:
        #     number_of_samples *= number_of_channels

        silent_numpy_array = np.zeros(
            shape = (number_of_channels, number_of_samples),
            dtype = dtype
        )

        return _AudioFrameUtils.from_numpy_audioframe(
            frame = silent_numpy_array,
            sample_rate = sample_rate,
            layout = layout,
            format = format,
            pts = pts,
            time_base = time_base
        )
    
    @staticmethod
    def from_numpy_audioframe(
        frame: np.ndarray,
        sample_rate: int,
        layout: str = FfmpegAudioLayout.STEREO.value,
        format: str = FfmpegAudioFormat.S16.value,
        pts: Union[int, None] = None,
        time_base: Union[Fraction, None] = None
    ) -> AudioFrame:
        """
        Transform the given numpy audio 'frame' into a
        pyav audio frame with the given 'sample_rate',
        'layout' and 'format, and also the 'pts
        and/or 'time_base' if provided.
        """
        layout = FfmpegAudioLayout.to_enum(layout).value
        format = FfmpegAudioFormat.to_enum(format).value

        frame = AudioFrame.from_ndarray(
            array = frame,
            format = format,
            layout = layout
        )

        frame.sample_rate = sample_rate

        if pts is not None:
            frame.pts = pts

        if time_base is not None:
            frame.time_base = time_base

        return frame
    
class _AudioUtils:
    """
    *For internal use only*

    Class to wrap some utils related to audio.
    """

    frame: _AudioFrameUtils = _AudioFrameUtils
    """
    Utils related to audio frames.
    """
    format: _AudioFormatUtils = _AudioFormatUtils
    """
    Utils related to audio format.
    """

    @staticmethod
    def iter_frames(
        filename: str
    ):
        """
        Iterator over the audio frames of the file
        with the given 'filename'.
        """
        container = av_open(file = filename)
        audio_stream = (
            container.streams.audio[0]
            if container.streams.audio else
            None
        )

        if audio_stream is None:
            raise Exception(f'No audio stream found in "{filename}" file.')
        
        for frame in _iter_frames(
            stream = audio_stream
        ):
            yield frame

        return
    
class VideoUtils:
    """
    Class to wrap information about a video, to
    detect alpha pixels, etc., and also the audio.
    """

    video: _VideoUtils = _VideoUtils
    """
    Utils related to video.
    """
    audio: _AudioUtils = _AudioUtils
    """
    Utils related to audio.
    """

def _iter_frames(
    stream: Union['VideoStream', 'AudioStream']
):
    """
    Iterator over the video frames of the video
    with the given 'filename'.
    """
    for packet in stream.container.demux(stream):
        for frame in packet.decode():
            yield frame

    # Some frames can be retained in the demuxer
    for frame in stream.decode():
        yield frame

    return